@@ -1,15 +1,15 @@
 """
 Provides tests for the syncing functionalities in django entity.
 """
+from django import db
 from django.contrib.contenttypes.models import ContentType
 from django.core.management import call_command
-from django.test.utils import override_settings
-from django_dynamic_fixture import G, F
+from django_dynamic_fixture import G
 from entity.config import EntityRegistry
-from entity.models import Entity, EntityRelationship
-from entity.sync import EntitySyncer
-from entity import turn_on_syncing, turn_off_syncing, sync_entities
-from mock import patch
+from entity.models import Entity, EntityRelationship, EntityKind
+from entity.sync import sync_entities, defer_entity_syncing, transaction_atomic_with_retry
+from entity.signal_handlers import turn_on_syncing, turn_off_syncing
+from mock import patch, MagicMock, call
 
 from entity.tests.models import (
     Account, Team, EntityPointer, DummyModel, MultiInheritEntity, AccountConfig, TeamConfig, TeamGroup,
@@ -22,10 +22,10 @@ class TestTurnOnOffSyncing(EntityTestCase):
     """
     Tests turning on and off entity syncing.
     """
-    @patch('entity.models.post_save', spec_set=True)
-    @patch('entity.models.post_delete', spec_set=True)
-    @patch('entity.models.m2m_changed', spec_set=True)
-    @patch('entity.models.post_bulk_operation', spec_set=True)
+    @patch('entity.signal_handlers.post_save', spec_set=True)
+    @patch('entity.signal_handlers.post_delete', spec_set=True)
+    @patch('entity.signal_handlers.m2m_changed', spec_set=True)
+    @patch('entity.signal_handlers.post_bulk_operation', spec_set=True)
     def test_turn_on_syncing_all_handlers_true(
             self, post_bulk_operation_mock, m2m_changed_mock, post_delete_mock, post_save_mock):
         turn_on_syncing(for_post_save=True, for_post_delete=True, for_m2m_changed=True, for_post_bulk_operation=True)
@@ -34,10 +34,10 @@ class TestTurnOnOffSyncing(EntityTestCase):
         self.assertTrue(m2m_changed_mock.connect.called)
         self.assertTrue(post_bulk_operation_mock.connect.called)
 
-    @patch('entity.models.post_save', spec_set=True)
-    @patch('entity.models.post_delete', spec_set=True)
-    @patch('entity.models.m2m_changed', spec_set=True)
-    @patch('entity.models.post_bulk_operation', spec_set=True)
+    @patch('entity.signal_handlers.post_save', spec_set=True)
+    @patch('entity.signal_handlers.post_delete', spec_set=True)
+    @patch('entity.signal_handlers.m2m_changed', spec_set=True)
+    @patch('entity.signal_handlers.post_bulk_operation', spec_set=True)
     def test_turn_on_syncing_all_handlers_false(
             self, post_bulk_operation_mock, m2m_changed_mock, post_delete_mock, post_save_mock):
         turn_on_syncing(
@@ -47,10 +47,10 @@ class TestTurnOnOffSyncing(EntityTestCase):
         self.assertFalse(m2m_changed_mock.connect.called)
         self.assertFalse(post_bulk_operation_mock.connect.called)
 
-    @patch('entity.models.post_save', spec_set=True)
-    @patch('entity.models.post_delete', spec_set=True)
-    @patch('entity.models.m2m_changed', spec_set=True)
-    @patch('entity.models.post_bulk_operation', spec_set=True)
+    @patch('entity.signal_handlers.post_save', spec_set=True)
+    @patch('entity.signal_handlers.post_delete', spec_set=True)
+    @patch('entity.signal_handlers.m2m_changed', spec_set=True)
+    @patch('entity.signal_handlers.post_bulk_operation', spec_set=True)
     def test_turn_off_syncing_all_handlers_true(
             self, post_bulk_operation_mock, m2m_changed_mock, post_delete_mock, post_save_mock):
         turn_off_syncing(for_post_save=True, for_post_delete=True, for_m2m_changed=True, for_post_bulk_operation=True)
@@ -59,10 +59,10 @@ class TestTurnOnOffSyncing(EntityTestCase):
         self.assertTrue(m2m_changed_mock.disconnect.called)
         self.assertTrue(post_bulk_operation_mock.disconnect.called)
 
-    @patch('entity.models.post_save', spec_set=True)
-    @patch('entity.models.post_delete', spec_set=True)
-    @patch('entity.models.m2m_changed', spec_set=True)
-    @patch('entity.models.post_bulk_operation', spec_set=True)
+    @patch('entity.signal_handlers.post_save', spec_set=True)
+    @patch('entity.signal_handlers.post_delete', spec_set=True)
+    @patch('entity.signal_handlers.m2m_changed', spec_set=True)
+    @patch('entity.signal_handlers.post_bulk_operation', spec_set=True)
     def test_turn_off_syncing_all_handlers_false(
             self, post_bulk_operation_mock, m2m_changed_mock, post_delete_mock, post_save_mock):
         turn_off_syncing(
@@ -76,7 +76,7 @@ class TestTurnOnOffSyncing(EntityTestCase):
         """
         Tests that save signals are connected by default.
         """
-        with patch('entity.models.sync_entities') as mock_handler:
+        with patch('entity.signal_handlers.sync_entities') as mock_handler:
             Account.objects.create()
             self.assertTrue(mock_handler.called)
 
@@ -94,7 +94,7 @@ class TestTurnOnOffSyncing(EntityTestCase):
         """
         Tests that bulk operations are turned off by default.
         """
-        with patch('entity.models.sync_entities') as mock_handler:
+        with patch('entity.signal_handlers.sync_entities') as mock_handler:
             Account.objects.bulk_create([Account() for i in range(5)])
             self.assertFalse(mock_handler.called)
 
@@ -103,7 +103,7 @@ class TestTurnOnOffSyncing(EntityTestCase):
         Tests turning off syncing for the save signal.
         """
         turn_off_syncing()
-        with patch('entity.models.sync_entities') as mock_handler:
+        with patch('entity.signal_handlers.sync_entities') as mock_handler:
             Account.objects.create()
             self.assertFalse(mock_handler.called)
 
@@ -112,7 +112,7 @@ class TestTurnOnOffSyncing(EntityTestCase):
         Tests turning off syncing for the delete signal.
         """
         turn_off_syncing()
-        with patch('entity.models.sync_entities') as mock_handler:
+        with patch('entity.signal_handlers.sync_entities') as mock_handler:
             a = Account.objects.create()
             self.assertFalse(mock_handler.called)
             a.delete()
@@ -123,7 +123,7 @@ class TestTurnOnOffSyncing(EntityTestCase):
         Tests turning off syncing for bulk operations.
         """
         turn_off_syncing()
-        with patch('entity.models.sync_entities') as mock_handler:
+        with patch('entity.signal_handlers.sync_entities') as mock_handler:
             Account.objects.bulk_create([Account() for i in range(5)])
             self.assertFalse(mock_handler.called)
 
@@ -133,7 +133,7 @@ class TestTurnOnOffSyncing(EntityTestCase):
         """
         turn_off_syncing()
         turn_on_syncing()
-        with patch('entity.models.sync_entities') as mock_handler:
+        with patch('entity.signal_handlers.sync_entities') as mock_handler:
             Account.objects.create()
             self.assertTrue(mock_handler.called)
 
@@ -154,7 +154,7 @@ class TestTurnOnOffSyncing(EntityTestCase):
         """
         turn_off_syncing()
         turn_on_syncing(for_post_bulk_operation=True)
-        with patch('entity.models.sync_entities') as mock_handler:
+        with patch('entity.signal_handlers.sync_entities') as mock_handler:
             Account.objects.bulk_create([Account() for i in range(5)])
             self.assertTrue(mock_handler.called)
 
@@ -179,22 +179,6 @@ class SyncAllEntitiesTest(EntityTestCase):
         call_command('sync_entities')
         self.assertEquals(Entity.objects.all().count(), 5)
 
-    @override_settings(CELERY_EAGER_PROPAGATES_EXCEPTIONS=True, CELERY_ALWAYS_EAGER=True, BROKER_BACKEND='memory')
-    def test_async_sync_entities_management_command(self):
-        """
-        Tests that the sync_entities command works with the asynchronous option.
-        """
-        # Create five test accounts without syncing on
-        turn_off_syncing()
-        for i in range(5):
-            Account.objects.create()
-        turn_on_syncing()
-
-        # Test that the management command syncs all five entities
-        self.assertEquals(Entity.objects.all().count(), 0)
-        call_command('sync_entities', async=True)
-        self.assertEquals(Entity.objects.all().count(), 5)
-
     def test_sync_dummy_data(self):
         """
         Tests that dummy data (i.e data that does not inherit EntityModelMixin) doesn't
@@ -266,6 +250,102 @@ class SyncAllEntitiesTest(EntityTestCase):
         # There should be four entity relationships since four accounts have teams
         self.assertEquals(EntityRelationship.objects.all().count(), 4)
 
+    def test_sync_all_accounts_teams_inactive_entity_kind(self):
+        """
+        Tests syncing of all accounts when they have super entities and the entiity kind is inactive
+        """
+        # Create five test accounts
+        accounts = [Account.objects.create() for i in range(5)]
+        # Create two teams to assign to some of the accounts
+        teams = [Team.objects.create() for i in range(2)]
+        accounts[0].team = teams[0]
+        accounts[0].save()
+        accounts[1].team = teams[0]
+        accounts[1].save()
+        accounts[2].team = teams[1]
+        accounts[2].save()
+        accounts[3].team = teams[1]
+        accounts[3].save()
+
+        team_ek = EntityKind.objects.get(name='tests.team')
+        team_ek.delete()
+
+        # Sync all the entities. There should be 7 (5 accounts 2 teams)
+        sync_entities()
+        self.assertEquals(Entity.objects.filter(entity_type=ContentType.objects.get_for_model(Account)).count(), 5)
+        self.assertEquals(Entity.objects.filter(entity_type=ContentType.objects.get_for_model(Team)).count(), 2)
+        self.assertEquals(Entity.objects.all().count(), 7)
+
+        # There should be four entity relationships since four accounts have teams
+        self.assertEquals(EntityRelationship.objects.all().count(), 4)
+
+
+class SyncSignalTests(EntityTestCase):
+    """
+    Test that when we are syncing we are calling the proper signals
+    """
+
+    @patch('entity.sync.model_activations_changed')
+    def test_sync_all(self, mock_model_activations_changed):
+        """
+        Tests that when we sync all we fire the correct signals
+        """
+
+        # Create five test accounts
+        turn_off_syncing()
+        initial_accounts = []
+        for i in range(5):
+            initial_accounts.append(Account.objects.create())
+        turn_on_syncing()
+
+        # Test that the management command syncs all five entities
+        self.assertEquals(Entity.objects.all().count(), 0)
+        sync_entities()
+        self.assertEquals(Entity.objects.all().count(), 5)
+        initial_entity_ids = list(Entity.objects.all().values_list('id', flat=True))
+        mock_model_activations_changed.send.assert_called_once_with(
+            sender=Entity,
+            instance_ids=sorted(initial_entity_ids),
+            is_active=True
+        )
+
+        # Create five new test accounts, and deactivate our initial accounts
+        mock_model_activations_changed.reset_mock()
+        turn_off_syncing()
+        new_accounts = []
+        for i in range(5):
+            new_accounts.append(Account.objects.create())
+        for account in initial_accounts:
+            account.delete()
+        turn_on_syncing()
+
+        # Sync entities
+        sync_entities()
+
+        # Assert that the correct signals were called
+        self.assertEqual(
+            mock_model_activations_changed.send.mock_calls,
+            [
+                call(
+                    sender=Entity,
+                    instance_ids=sorted(list(Entity.objects.filter(
+                        entity_id__in=[account.id for account in new_accounts]
+                    ).values_list('id', flat=True))),
+                    is_active=True
+                ),
+                call(
+                    sender=Entity,
+                    instance_ids=sorted(initial_entity_ids),
+                    is_active=False
+                )
+            ]
+        )
+
+        # Test syncing all when nothing should have changed
+        mock_model_activations_changed.reset_mock()
+        sync_entities()
+        self.assertFalse(mock_model_activations_changed.send.called)
+
 
 class TestEntityBulkSignalSync(EntityTestCase):
     """
@@ -778,21 +858,6 @@ class TestCachingAndCascading(EntityTestCase):
         self.assertEquals(Entity.objects.count(), 2)
         self.assertEquals(EntityRelationship.objects.count(), 1)
 
-    def test_no_cascade_if_super_entity_exists(self):
-        """
-        Tests that super entities arent synced again if they have already been synced.
-        """
-        account = G(Account, team=F(team_group=F()))
-        self.assertTrue(Account.objects.exists())
-        self.assertTrue(Team.objects.exists())
-        self.assertTrue(TeamGroup.objects.exists())
-
-        entity_syncer = EntitySyncer()
-        entity_syncer.sync_entities_and_relationships(account)
-        # Verify that only the account and team reside in the entity syncer cache. This means that
-        # the syncing didnt percolate all the way to the team group
-        self.assertEquals(len(entity_syncer._synced_entity_cache), 2)
-
     def test_optimal_queries_registered_entity_with_no_qset(self):
         """
         Tests that the optimal number of queries are performed when syncing a single entity that
@@ -801,7 +866,7 @@ class TestCachingAndCascading(EntityTestCase):
         team_group = G(TeamGroup)
 
         ContentType.objects.clear_cache()
-        with self.assertNumQueries(6):
+        with self.assertNumQueries(15):
             team_group.save()
 
     def test_optimal_queries_registered_entity_w_qset(self):
@@ -811,7 +876,7 @@ class TestCachingAndCascading(EntityTestCase):
         account = G(Account)
 
         ContentType.objects.clear_cache()
-        with self.assertNumQueries(7):
+        with self.assertNumQueries(18):
             account.save()
 
     def test_sync_all_optimal_queries(self):
@@ -834,15 +899,13 @@ class TestCachingAndCascading(EntityTestCase):
         # Use an entity registry that only has accounts and teams. This ensures that other registered
         # entity models dont pollute the test case
         new_registry = EntityRegistry()
-        new_registry.register_entity(
-            Account.objects.select_related('team', 'team2', 'team_group', 'competitor'), AccountConfig)
-        new_registry.register_entity(
-            Team.objects.select_related('team_group'), TeamConfig)
+        new_registry.register_entity(AccountConfig)
+        new_registry.register_entity(TeamConfig)
 
         with patch('entity.sync.entity_registry') as mock_entity_registry:
             mock_entity_registry.entity_registry = new_registry.entity_registry
             ContentType.objects.clear_cache()
-            with self.assertNumQueries(27):
+            with self.assertNumQueries(19):
                 sync_entities()
 
         self.assertEquals(Entity.objects.filter(entity_type=ContentType.objects.get_for_model(Account)).count(), 5)
@@ -851,3 +914,95 @@ class TestCachingAndCascading(EntityTestCase):
 
         # There should be four entity relationships since four accounts have teams
         self.assertEquals(EntityRelationship.objects.all().count(), 4)
+
+
+class DeferEntitySyncingTests(EntityTestCase):
+    """
+    Tests the defer entity syncing decorator
+    """
+
+    def test_defer(self):
+        @defer_entity_syncing
+        def test_method(test, count=5, sync_all=False):
+            # Create some entities
+            for i in range(count):
+                Account.objects.create()
+
+            if sync_all:
+                sync_entities()
+
+            # Assert that we do not have any entities
+            test.assertEquals(Entity.objects.all().count(), 0)
+
+        # Call the test method
+        test_method(self, count=5)
+
+        # Assert that after the method was run we did sync the entities
+        self.assertEquals(Entity.objects.all().count(), 5)
+
+        # Delete all entities
+        Entity.all_objects.all()._raw_delete(Entity.objects.db)
+
+        # Call the method again syncing all
+        test_method(self, count=0, sync_all=True)
+
+        # Assert that after the method was run we did sync the entities
+        self.assertEquals(Entity.objects.all().count(), 5)
+
+        # Assert that we restored the defer flag
+        self.assertFalse(sync_entities.defer)
+
+        # Assert that we cleared the buffer
+        self.assertEqual(sync_entities.buffer, {})
+
+    def test_defer_nothing_synced(self):
+        """
+        Test the defer decorator when nothing is synced
+        """
+
+        @defer_entity_syncing
+        def test_method(test):
+            # Assert that we do not have any entities
+            test.assertEquals(Entity.objects.all().count(), 0)
+
+        # Call the test method
+        with patch('entity.sync.sync_entities') as mock_sync_entities:
+            # Call the method that does no syncing
+            test_method(self)
+
+            # Ensure that we did not call sync entities
+            self.assertFalse(mock_sync_entities.called)
+
+
+class TransactionAtomicWithRetryTests(EntityTestCase):
+    """
+    Test the transaction_atomic_with_retry decorator
+    """
+
+    def test_retry_operational_error(self):
+        exception_mock = MagicMock()
+        exception_mock.side_effect = db.utils.OperationalError()
+
+        @transaction_atomic_with_retry()
+        def test_func():
+            exception_mock()
+
+        with self.assertRaises(db.utils.OperationalError):
+            test_func()
+
+        self.assertEqual(
+            len(exception_mock.mock_calls),
+            6
+        )
+
+    def test_retry_other_error(self):
+        exception_mock = MagicMock()
+        exception_mock.side_effect = Exception()
+
+        @transaction_atomic_with_retry()
+        def test_func():
+            exception_mock()
+
+        with self.assertRaises(Exception):
+            test_func()
+        exception_mock.assert_called_once_with()
